import subprocess
from pathlib import Path
from sys import stderr

from termcolor import colored


def install_node_packages(node_prefix: Path | None = None) -> None:
    node_prefix = node_prefix or Path(__file__).parent
    if not (node_prefix / 'node_modules').exists():
        print(colored('Installing Node.js packages', 'yellow', attrs=['bold']), file=stderr)
        command = ['npm', 'install', '--no-save', '--prefix', str(node_prefix)]
        subprocess.run(command, text=True, check=True)  # nosec: secure, not using untrusted input



import re
import subprocess

import black
import pytest
from logikal_utils.project import PYPROJECT

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def get_mode(max_line_length: int) -> black.Mode:
    config = PYPROJECT.get('tool', {}).get('black', {})
    return black.Mode(
        line_length=config.get('line-length', max_line_length),
        string_normalization=not config.get('skip-string-normalization', True),
        preview=config.get('preview', True),
    )


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('black')
    group.addoption('--black', action='store_true', default=False, help='run black')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'black: tests checking code formatting.')
    if config.option.black:
        config.pluginmanager.register(BlackPlugin(config=config))


class BlackItem(CachedFileCheckItem):
    def run(self) -> None:
        mode = get_mode(max_line_length=int(self.config.getini('max_line_length')))
        command = [
            'black', '--check', '--diff', '--color', '--quiet', str(self.path),
            '--line-length', str(mode.line_length),
        ]
        if not mode.string_normalization:
            command += ['--skip-string-normalization']
        if mode.preview:
            command += ['--preview']

        process = subprocess.run(command, capture_output=True, text=True, check=False)  # nosec
        if process.returncode:
            message = re.sub(r'^.*\n.*\n([^@]*)@@', r'\1@@', process.stdout) or process.stderr
            raise ItemRunError(message.lstrip('\n').rstrip())


class BlackPlugin(CachedFileCheckPlugin):
    name = 'black'
    item = BlackItem



from collections.abc import Callable, Generator
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, TypeVar

from logikal_utils.project import PYPROJECT

from pytest_logikal.core import DEFAULT_INI_OPTIONS

logger = getLogger(__name__)

Function = TypeVar('Function', bound=Callable[..., Any])
Fixture = Callable[[Function], Function]


def get_ini_option(name: str) -> Any:
    ini_options = PYPROJECT.get('tool', {}).get('pytest', {}).get('ini_options', {})
    default = DEFAULT_INI_OPTIONS[name]['value']
    return type(default)(ini_options.get(name, default))


@contextmanager
def render_template(path: Path, context: dict[str, Any]) -> Generator[Path]:
    with TemporaryDirectory(prefix='pytest_logikal_', suffix='_template') as tmp_dir:
        rendered = path.read_text(encoding='utf-8').format(**context)
        config_path = Path(tmp_dir) / path.name
        config_path.write_text(rendered)
        yield config_path



import logging
import os
import shutil
import sys
import warnings
from collections.abc import Callable, Iterator
from importlib.util import find_spec
from itertools import chain
from pathlib import Path
from typing import Any

import pytest
from pytest_cov import CovFailUnderWarning
from termcolor import colored

sys.path.insert(0, os.getcwd())
warnings.filterwarnings(action='ignore', category=CovFailUnderWarning)

PLUGINS = {
    'core': [
        'mypy', 'bandit', 'build', 'docs', 'isort', 'licenses', 'pylint', 'requirements', 'spell',
        'style',
    ],
    'django': ['migration', 'translations', 'html', 'css', 'svg', 'js'],
}
DEFAULT_INI_OPTIONS: dict[str, Any] = {
    'max_line_length': {'value': 99, 'help': 'the maximum line length to use'},
    'max_complexity': {'value': 10, 'help': 'the maximum complexity to allow'},
    'cov_fail_under': {'value': 100, 'help': 'target coverage percentage'},
}
EXTRAS = {
    'black': bool(find_spec('black')),
    'browser': bool(find_spec('logikal_browser')),
    'django': bool(find_spec('pytest_django')),
}

ReportInfoType = tuple[Any | str, int | None, str]


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('logikal')
    group.addoption('--live', action='store_true', help='run live tests')
    group.addoption('--fast', action='store_true', help='do not run any additional checks')
    group.addoption('--clear', action='store_true', help='clear cache before running tests')
    group.addoption('--no-defaults', action='store_true', help='do not use our own defaults')
    group.addoption('--no-mypy', action='store_true', help='do not use mypy')
    group.addoption('--no-bandit', action='store_true', help='do not use bandit')
    group.addoption('--no-build', action='store_true', help='do not run build checks')
    group.addoption('--no-docs', action='store_true', help='do not run documentation checks')
    group.addoption('--no-isort', action='store_true', help='do not use isort')
    group.addoption('--no-licenses', action='store_true', help='do not check licenses')
    group.addoption('--no-pylint', action='store_true', help='do not use pylint')
    group.addoption('--no-requirements', action='store_true', help='do not check requirements')
    group.addoption('--no-spell', action='store_true', help='do not use codespell')
    group.addoption('--no-style', action='store_true', help='do not use pycodestyle & pydocstyle')
    group.addoption('--no-install', action='store_true', help='do not install packages')

    if EXTRAS['django']:
        group.addoption('--no-migration', action='store_true', help='do not check migrations')
        group.addoption('--no-translations', action='store_true', help='do not check translations')
        group.addoption('--no-html', action='store_true', help='do not run html template checks')
        group.addoption('--no-css', action='store_true', help='do not run css checks')
        group.addoption('--no-svg', action='store_true', help='do not run svg checks')
        group.addoption('--no-js', action='store_true', help='do not run js checks')

    for option, entry in DEFAULT_INI_OPTIONS.items():
        parser.addini(option, default=str(entry['value']), help=entry['help'])


def pytest_addhooks(pluginmanager: pytest.PytestPluginManager) -> None:
    # Block extra plugins that are not installed
    for extra, installed in EXTRAS.items():
        if not installed:
            pluginmanager.set_blocked(f'logikal_{extra}')
            for plugin in PLUGINS.get(extra, [extra]):
                pluginmanager.set_blocked(f'logikal_{plugin}')


@pytest.hookimpl(wrapper=True)
def pytest_load_initial_conftests(
    early_config: pytest.Config, args: list[str],
) -> Iterator[None]:
    if '--no-defaults' in args:
        yield

    # Updating arguments
    args.extend(['--strict-config', '--strict-markers'])
    namespace = early_config.known_args_namespace
    if '--live' in args:
        args.append('--capture=no')
        namespace.capture = 'no'
    if '--clear' in args:
        args.append('--cache-clear')
        namespace.cacheclear = True
    if '-r' not in args:
        args.extend(['-r', 'fExX'])
    if '-n' not in args:
        args.extend(['-n', 'auto' if '--live' not in args else '0'])
        namespace.dist = 'load' if '--live' not in args else 'no'
    for plugin in chain(*(plugins for name, plugins in PLUGINS.items() if EXTRAS.get(name, True))):
        if '--fast' not in args and f'--no-{plugin}' not in args:
            args.append(f'--{plugin}')
            setattr(namespace, plugin, True)
    if all(arg not in args for arg in ['--cov', '--no-cov', '--live', '--fast']):
        args.extend(['--cov', '--no-cov-on-fail'])
        namespace.no_cov_on_fail = True

    # Setting coverage options
    if '--cov' in args:
        fail_under = early_config.getini('cov_fail_under')
        namespace.cov_source = ['.']
        namespace.cov_fail_under = float(fail_under) if '.' in fail_under else int(fail_under)
        namespace.cov_report = {'term-missing': 'skip-covered'}

    # Updating config
    ini_defaults: dict[str, str | list[str]] = {
        'console_output_style': 'classic',
        'xfail_strict': 'True',
        'filterwarnings': ['error'],
        'log_cli': 'True' if '--live' in args else 'False',
        'log_level': 'DEBUG',
        'log_format': '%(asctime)s.%(msecs)03d %(levelname)s %(message)s (%(name)s:%(lineno)s)',
        'log_date_format': '%Y-%m-%d %H:%M:%S',
        'log_auto_indent': 'True',
    }
    early_config.inicfg = {**ini_defaults, **early_config.inicfg}
    yield


def pytest_sessionstart(session: pytest.Session) -> None:
    # Clearing cache
    if session.config.getoption('clear'):
        print(colored('Clearing cache', 'yellow', attrs=['bold']))
        Path('.coverage').unlink(missing_ok=True)
        shutil.rmtree('.pytest_cache', ignore_errors=True)
        shutil.rmtree('.mypy_cache', ignore_errors=True)


def pytest_configure(config: pytest.Config) -> None:
    # Hiding information
    if not config.getoption('verbose'):
        # Hiding overly verbose debug and info log messages
        logging.getLogger('django_migration_linter').setLevel(logging.WARNING)
        logging.getLogger('docker').setLevel(logging.INFO)
        logging.getLogger('faker').setLevel(logging.INFO)
        logging.getLogger('filelock').setLevel(logging.WARNING)
        logging.getLogger('matplotlib').setLevel(logging.INFO)
        logging.getLogger('PIL').setLevel(logging.INFO)
        logging.getLogger('pydocstyle').setLevel(logging.WARNING)
        logging.getLogger('pytest_logikal.validator').setLevel(logging.INFO)
        logging.getLogger('urllib3').setLevel(logging.INFO)
        logging.getLogger('pydocstyle').setLevel(logging.ERROR)
        logging.getLogger('blib2to3').setLevel(logging.WARNING)  # used by Black

        # Hiding worker information lines
        reports = config.pluginmanager.get_plugin('reports')
        reports.getworkerinfoline = lambda *_args, **_kwargs: ''  # type: ignore[union-attr]

    # Configuring mypy
    mypy = config.pluginmanager.get_plugin('mypy')
    mypy.mypy_argv = [  # type: ignore[union-attr]
        '--no-error-summary',
        '--strict',
        '--show-column-numbers',
        '--warn-unreachable',
        '--local-partial-types',
    ]


def unified_reportinfo(item: pytest.Item, header: str) -> Callable[[], ReportInfoType]:
    def reportinfo() -> ReportInfoType:
        return (item.path, None, header)
    return reportinfo


@pytest.hookimpl(trylast=True)
def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    for item in items:
        source = item.nodeid.split('::')[-1].lower()
        if source == 'mypy':
            header = f'[{item.name}] {item.path.relative_to(item.config.invocation_params.dir)}'
            setattr(item, 'reportinfo', unified_reportinfo(item=item, header=header))
        if source == 'mypy-status':
            setattr(item, '_nodeid', '::check::mypy-status')
            setattr(item, 'reportinfo', unified_reportinfo(item=item, header='[mypy-status]'))



import json
import re
import subprocess

import pytest
from logikal_utils.project import PYPROJECT

from pytest_logikal.core import ReportInfoType
from pytest_logikal.plugin import Item, ItemRunError, Plugin

ALLOWED_LICENSES = [
    r'^3-Clause BSD License$',
    r'^Apache 2\.0$',
    r'^Apache License 2\.0$',
    r'^Apache License Version 2\.0$',
    r'^Apache License, Version 2\.0$',
    r'^Apache Software License$',
    r'^BSD 3-Clause$',
    r'^BSD License$',
    r'^BSD$',
    r'^BSD-3-Clause$',
    r'^CC0 1.0 Universal \(CC0 1\.0\) Public Domain Dedication$',
    r'^CMU License \(MIT-CMU\)$',
    r'^GNU Lesser General Public License v2 \(LGPLv2\)$',
    r'^GNU Lesser General Public License v2 or later \(LGPLv2\+\)$',
    r'^GNU Lesser General Public License v3 \(LGPLv3\)$',
    r'^GNU Library or Lesser General Public License \(LGPL\)$',
    r'^ISC License \(ISCL\)$',
    r'^ISC$',
    r'^MIT License$',
    r'^MIT No Attribution License \(MIT-0\)$',
    r'^MIT$',
    r'^Mozilla Public License 2\.0 \(MPL 2\.0\)$',
    r'^Public Domain$',
    r'^Python Software Foundation License$',
    r'^The MIT License \(MIT\)$',
    r'^The Unlicense \(Unlicense\)$',
]

ALLOWED_PACKAGES = {
    'codespell': r'^GPL-2\.0-only$',  # only used as a local tool
    'django-migration-linter': r'^UNKNOWN$',  # license is Apache License 2.0, see [1]
    'djlint': r'^GNU General Public License v3 or later \(GPLv3\+\)$',  # only used as a local tool
    'facebook-business': r'^LICENSE\.txt$',  # only used as a connector
    'facebook_business': r'^LICENSE\.txt$',  # only used as a connector
    'html-tag-names': r'^GNU General Public License v3 or later \(GPLv3\+\)$',  # local tool
    'html-void-elements': r'^GNU General Public License v3 or later \(GPLv3\+\)$',  # local tool
    'pkg-resources': r'^UNKNOWN$',  # caused by an Ubuntu bug, see [2]
    'pkg_resources': r'^UNKNOWN$',  # caused by an Ubuntu bug, see [2]
    'pylint': r'^GNU General Public License v2 \(GPLv2\)$',  # only used as a local tool
    'pylint-django': r'^GNU General Public License v2 or later \(GPLv2\+\)$',  # local plugin
    'pylint-plugin-utils': r'^GNU General Public License v2 or later \(GPLv2\+\)$',  # local plugin
    # Broken license inference because of PEP 639, see [3]
    'attrs': r'^UNKNOWN$',  # license is MIT
    'prettytable': r'^UNKNOWN$',  # license is BSD-3-Clause
    'referencing': r'^UNKNOWN$',  # license is MIT
    'jeepney': r'^UNKNOWN$',  # license is MIT
    'Sphinx': r'^UNKNOWN$',  # license is BSD-2-Clause
}
# [1] https://github.com/3YOURMIND/django-migration-linter/issues/290
# [2] https://bugs.launchpad.net/ubuntu/+source/python-pip/+bug/1635463
# [3] https://github.com/raimon49/pip-licenses/issues/225


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('licenses')
    group.addoption('--licenses', action='store_true', default=False, help='run license checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'licenses: checks licenses.')
    if config.option.licenses:
        config.pluginmanager.register(LicensePlugin(config=config))


class LicenseItem(Item):
    def runtest(self) -> None:
        command = ['pip-licenses', '--format=json', '--with-system', '--with-urls']

        # This subprocess call is secure as it is not using untrusted input
        process = subprocess.run(command, capture_output=True, text=True, check=True)  # nosec
        packages = json.loads(process.stdout)

        pyproject_licenses = PYPROJECT.get('tool', {}).get('licenses', {})
        allowed_licenses = pyproject_licenses.get('allowed_licenses', ALLOWED_LICENSES)
        allowed_packages = pyproject_licenses.get('allowed_packages', ALLOWED_PACKAGES)
        allowed_licenses.extend(pyproject_licenses.get('extend_allowed_licenses', []))
        allowed_packages.update(pyproject_licenses.get('extend_allowed_packages', {}))

        warnings: list[str] = []
        for package in sorted(packages, key=lambda item: (item['License'], item['Name'].lower())):
            name, version, url = package['Name'], package.get('Version'), package.get('URL')
            licenses = [license.strip() for license in package['License'].split(';')]

            allowed_package_license = any(
                re.match(allowed_packages[name], license)
                for license in licenses
            ) if name in allowed_packages else False
            allowed_license = any(
                re.match(allowed_license, license)
                for license in licenses
                for allowed_license in allowed_licenses
            )

            if (not allowed_package_license and not allowed_license):
                url = f' ({url})' if url not in ('UNKNOWN', None) else ''
                warnings.append(f'Warning: {package['License']}: {name}=={version}{url}')
        if warnings:
            raise ItemRunError('\n'.join(warnings))

    def reportinfo(self) -> ReportInfoType:
        return (self.path, None, f'[{LicensePlugin.name}]')


class LicensePlugin(Plugin):
    name = 'licenses'
    item = LicenseItem
